<?php

namespace LakeDrops\Component\Composer;

use Composer\Json\JsonFile;
use Exception;
use Symfony\Component\Yaml\Yaml;
use Twig\Error\LoaderError;
use Twig\Error\RuntimeError;
use Twig\Error\SyntaxError;
use Twig_Environment;
use Twig_Loader_Array;

/**
 * Manages project config.
 */
final class Config {

  /**
   * @var array
   */
  private static $values;

  /**
   * @var array
   */
  private static $customValues;

  /**
   * @var array
   */
  private static $overwriteValues;

  /**
   * @var array
   */
  private static $userValues;

  /**
   * @var string
   */
  private $component;

  /**
   * @var \Twig_Loader_Array
   */
  private $twig_loader;

  /**
   * @var \Twig_Environment
   */
  private $twig;

  /**
   * @var \LakeDrops\Component\Composer\Dotenv
   */
  private $env;

  /**
   * @var string
   */
  private $configFile;

  /**
   * @var string
   */
  private $userConfigFile;

  /**
   * Config constructor.
   *
   * @param string $component
   * @param array $default_values
   * @param \LakeDrops\Component\Composer\Dotenv $env
   */
  public function __construct(string $component, array $default_values, Dotenv $env) {
    $this->component = $component;
    $this->env = $env;
    $this->twig_loader = new Twig_Loader_Array([]);
    $this->twig = new Twig_Environment($this->twig_loader);
    $this->configFile = getcwd() . '/.lakedrops.yml';
    $this->userConfigFile = getcwd() . '/.lakedrops.user.yml';

    $this->init();
    $this->merge($default_values, self::$customValues[$this->component] ?? []);
    $this->merge([], self::$overwriteValues[$this->component] ?? [], FALSE);
    $this->merge([], self::$userValues[$this->component] ?? [], FALSE);
    $this->migrateFromComposerJson();
  }

  /**
   * Read project configuration if not already happened.
   */
  private function init(): void {
    if (self::$values === NULL) {
      self::$values = [];
      self::$overwriteValues = [];
      if (!file_exists($this->configFile)) {
        self::$customValues = [];
        $this->save();
      }
      else {
        self::$customValues = Yaml::parseFile($this->configFile);
        // Apply overwrite by stage declarations.
        $stage = getenv('PROJECT_BRANCH');
        if (isset(self::$customValues['stage_overwrites'][$stage])) {
          self::$overwriteValues = self::$customValues['stage_overwrites'][$stage];
        }
      }
      if (!file_exists($this->userConfigFile)) {
        self::$userValues = [];
      }
      else {
        self::$userValues = Yaml::parseFile($this->userConfigFile);
      }
    }
  }

  /**
   * Migrate custom settings from composer.json if required.
   */
  private function migrateFromComposerJson(): void {
    $filename = getcwd() . '/composer.json';
    $jsonFile = new JsonFile($filename);
    $content = $jsonFile->read();
    if (isset($content['extra'][$this->component])) {
      $this->merge([], $content['extra'][$this->component]);
      $this->save();
      unset($content['extra'][$this->component]);
      try {
        $jsonFile->write($content);
      } catch (Exception $ex) {
        // Let's ignore this, if composer.json is read-only there is a general problem.
      }
    }
  }

  /**
   * @param array $defaultValues
   * @param array $customValues
   * @param bool $store
   */
  private function merge(array $defaultValues, array $customValues, $store = TRUE): void {
    $default = self::$values[$this->component] ?? [];
    $custom = self::$customValues[$this->component] ?? [];
    if (!empty($defaultValues)) {
      $default = NestedArray::mergeDeep($default, $defaultValues);
    }
    if (!empty($customValues)) {
      $default = NestedArray::mergeDeep($default, $customValues);
      if ($store) {
        $custom = NestedArray::mergeDeep($custom, $customValues);
      }
    }
    self::$values[$this->component] = $this->env->replaceEnvironmentVariables($default);
    self::$customValues[$this->component] = $custom;
  }

  /**
   * Save the current settings.
   */
  private function save(): void {
    file_put_contents($this->configFile, Yaml::dump(self::$customValues, 9, 2));
  }

  /**
   * @param array $values
   * @param array $keys
   *
   * @return mixed|null
   */
  private function readValueFromArray(array $values, array $keys) {
    $key = array_shift($keys);
    if (!isset($values[$key])) {
      return NULL;
    }
    if (empty($keys)) {
      return $values[$key];
    }
    return $this->readValueFromArray($values[$key], $keys);
  }

  /**
   * @param string|array $keys
   *
   * @return mixed|null
   */
  public function readValue($keys) {
    if (is_string($keys)) {
      $keys = [$keys];
    }
    return $this->readValueFromArray(self::$values[$this->component], $keys);
  }

  /**
   * @param string $key
   * @param mixed $value
   * @param bool $store
   */
  public function setValue(string $key, $value, $store = TRUE): void {
    $this->merge([], [$key => $value], $store);
    $this->save();
  }

  /**
   * @param string $filename
   * @param string $content
   *
   * @return string
   */
  public function render(string $filename, string $content): string {
    $this->twig_loader->setTemplate($filename, $content);
    try {
      return $this->twig->render($filename, self::$values[$this->component]);
    }
    catch (LoaderError $e) {
    }
    catch (RuntimeError $e) {
    }
    catch (SyntaxError $e) {
    }
    return '';
  }

}
